# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import re
import numpy
import instant
import hashlib
import types

import gotran
from gotran.common import (
    check_arg,
    check_kwarg,
    push_log_level,
    pop_log_level,
    info,
    INFO,
    value_error,
)
from gotran.model.ode import ODE
from gotran.model.loadmodel import load_ode
from gotran.common.options import parameters
from gotran.codegeneration.codegenerators import (
    PythonCodeGenerator,
    CCodeGenerator,
    class_name,
    DOLFINCodeGenerator,
)
from gotran.codegeneration.algorithmcomponents import (
    rhs_expressions,
    monitored_expressions,
    jacobian_expressions,
)

# Set log level of instant
instant.set_log_level("WARNING")

__all__ = ["compile_module"]

additional_declarations_ = r"""
%init%{{
import_array();
%}}

%include <exception.i>
%feature("autodoc", "1");

// Typemaps
%typemap(in) (double* {rhs_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles "
                  "expected. Make sure the numpy array is contiguous, "
                  "and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                                  "{num_states}, for the {rhs_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (const double* {states_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                              "{num_states}, for the {states_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (double* {states_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                              "{num_states}, for the {states_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (double* {parameters_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_parameters} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
             "{num_parameters}, for the {parameters_name} argument.");

  $1 = (double *)PyArray_DATA(xa);

}}

// The typecheck
%typecheck(SWIG_TYPECHECK_DOUBLE_ARRAY) double *
{{
    $1 = PyArray_Check($input) ? 1 : 0;
}}

%typecheck(SWIG_TYPECHECK_DOUBLE_ARRAY) const double *
{{
    $1 = PyArray_Check($input) ? 1 : 0;
}}

%pythoncode%{{
def {rhs_function_name}({args},{rhs_name} = None):
    '''
    Evaluates the right hand side of the model

    Arguments
    ---------
{args_doc}
    {rhs_name} : np.ndarray (optional)
        The computed result
    '''
    import numpy as np
    if {rhs_name} is None:
        {rhs_name} = np.zeros_like({states_name})

    _{rhs_function_name}({args}, {rhs_name})
    return {rhs_name}

{python_code}
%}}

%rename (_{rhs_function_name}) {rhs_function_name};

{jacobian_declaration}

{monitor_declaration}
"""

jacobian_declaration_template_ = """
// Typemaps
%typemap(in) (double* {jac_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles "
                  "expected. Make sure the numpy array is contiguous, "
                  "and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states}*{num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                                  "{num_states}*{num_states}, for the {jac_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%pythoncode%{{
def {jacobian_function_name}({args}, {jac_name}=None):
    '''
    Evaluates the jacobian of the model

    Arguments
    ---------
{args_doc}
    {jac_name} : np.ndarray (optional)
        The computed result
    '''
    import numpy as np
    if {jac_name} is None:
        {jac_name} = np.zeros({num_states}*{num_states}, dtype=np.float_)
    elif not isinstance({jac_name}, np.ndarray):
        raise TypeError(\"expected a NumPy array.\")
    elif len({jac_name}.shape) != 2 or {jac_name}.shape[0] != {jac_name}.shape[1] or {jac_name}.shape[0] != {num_states}:
        raise ValueError(\"expected a square shaped matrix with size ({num_states}, {num_states})\")
    else:
        # Flatten Matrix
        {jac_name}.shape = ({num_states}*{num_states},)

    _{jacobian_function_name}({args}, {jac_name})
    {jac_name}.shape = ({num_states},{num_states})
    return {jac_name}
%}}

%rename (_{jacobian_function_name}) {jacobian_function_name};
"""

monitor_declaration_template = """
%typemap(in) (double* {monitored_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_monitored} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
             "{num_monitored}, for the {monitored_name} argument.");

  $1 = (double *)PyArray_DATA(xa);

}}

%pythoncode%{{
def {monitored_function_name}({args}, {monitored_name}=None):
    '''
    Evaluates any monitored intermediates of the model

    Arguments
    ---------
{args_doc}
    {monitored_name} : np.ndarray (optional)
        The computed result
    '''
    import numpy as np
    if {monitored_name} is None:
        {monitored_name} = np.zeros({num_monitored}, dtype=np.float_)
    elif not isinstance({monitored_name}, np.ndarray):
        raise TypeError(\"expected a NumPy array.\")
    elif len({monitored_name}) != {num_monitored}:
        raise ValueError(\"expected a numpy array of size: {num_monitored}\")

    _{monitored_function_name}({args}, {monitored_name})
    return {monitored_name}
%}}

%rename (_{monitored_function_name}) {monitored_function_name};
"""


def compile_module(
    ode,
    language="C",
    monitored=None,
    generation_params=None,
    additional_declarations=None,
    jacobian_declaration_template=None,
):
    """
    JIT compile an ode

    Arguments
    ---------
    ode : gotran.ODE, str
        The gotran ode
    language : str (optional)
        The language of the generated code
        Defaults : 'C' ['C', 'Python']
    monitored : list
        A list of names of intermediates of the ODE. Code for monitoring
        the intermediates will be generated.
    generation_params : dict
        Parameters controling the code generation
    """

    monitored = monitored or []
    generation_params = generation_params or {}

    check_arg(ode, (ODE, str))

    if isinstance(ode, str):
        ode = load_ode(ode)

    check_kwarg(language, "language", str)

    language = language.capitalize()
    valid_languages = ["C", "Python", "Dolfin"]
    if language not in valid_languages:
        value_error(
            "Expected one of {0} for the language kwarg.".format(
                ", ".join("'{0}'".format(lang) for lang in valid_languages)
            )
        )

    params = parameters.generation.copy()
    params.update(generation_params)

    if language == "C":
        return compile_extension_module(
            ode,
            monitored,
            params,
            additional_declarations,
            jacobian_declaration_template,
        )

    # Create unique module name for this application run
    modulename = "gotran_{0}_module_{1}_{2}".format(
        language.lower(),
        class_name(ode.name),
        hashlib.sha1(
            str(
                ode.signature()
                + str(monitored)
                + repr(params)
                + gotran.__version__
                + instant.__version__
            ).encode("utf-8")
        ).hexdigest(),
    )

    # Check cache
    python_module = instant.import_module(modulename)
    if python_module:
        return getattr(python_module, class_name(ode.name))()

    # No module in cache generate python version
    if language == "Dolfin":
        pgen = DOLFINCodeGenerator(params)
    else:
        pgen = PythonCodeGenerator(params)

    # Generate class code, execute it and collect namespace
    code = (
        "from __future__ import division\nimport numpy as np\nimport math"
        + pgen.class_code(ode, monitored=monitored)
    )

    # Make a temporary module path for compilation
    module_path = os.path.join(instant.get_temp_dir(), modulename)
    instant.instant_assert(
        not os.path.exists(module_path),
        "Not expecting module_path to exist: '{}'".format(module_path),
    )
    instant.makedirs(module_path)
    original_path = os.getcwd()
    try:
        module_path = os.path.abspath(module_path)
        os.chdir(module_path)
        instant.write_file("__init__.py", code)
        module_path = instant.copy_to_cache(
            module_path, instant.get_default_cache_dir(), modulename
        )

    finally:
        # Always get back to original directory.
        os.chdir(original_path)

    python_module = instant.import_module(modulename)
    return getattr(python_module, class_name(ode.name))()


def compile_extension_module(
    ode,
    monitored,
    params,
    additional_declarations=None,
    jacobian_declaration_template=None,
):
    """
    Compile an extension module, based on the C code from the ode
    """

    # Add function prototype
    args = []
    args_doc = []
    for arg in params.code.default_arguments:
        if arg == "s":
            args.append("states")
            args_doc.append(
                """    {0} : np.ndarray
        The state values""".format(
                    params.code.states.array_name
                )
            )
        elif arg == "t":
            args.append("time")
            args_doc.append(
                """    time : scalar
        The present time"""
            )
        elif arg == "p" and params.code.parameters.representation != "numerals":
            args.append("parameters")
            args_doc.append(
                """    {0} : np.ndarray
        The parameter values""".format(
                    params.code.parameters.array_name
                )
            )
        elif arg == "b" and params.code.body.representation != "named":
            args.append("body")
            args_doc.append(
                """    {0} : np.ndarray
        The body values""".format(
                    params.code.body.array_name
                )
            )

    # Create unique module name for this application run
    modulename = "gotran_compiled_module_{0}_{1}".format(
        class_name(ode.name),
        hashlib.sha1(
            str(
                ode.signature()
                + str(monitored)
                + repr(params)
                + gotran.__version__
                + instant.__version__
            ).encode("utf-8")
        ).hexdigest(),
    )

    # Check cache
    compiled_module = instant.import_module(modulename)

    if compiled_module:
        return compiled_module

    args = ", ".join(args)
    args_doc = "\n".join(args_doc)

    # Do not generate any Python functions
    python_params = params.copy()
    for name in python_params.functions:
        if name == "monitored":
            continue
        python_params.functions[name].generate = False

    jacobian_declaration = ""
    monitor_declaration = ""
    if params.functions.jacobian.generate:

        # Flatten jacobian params
        if not params.code.array.flatten:
            debug(
                "Generating jacobian C-code, forcing jacobian array " "to be flattened."
            )
            params.code.array.flatten = True

        jacobian_declaration_template = (
            jacobian_declaration_template_
            if jacobian_declaration_template is None
            else jacobian_declaration_template
        )

        jacobian_declaration = jacobian_declaration_template.format(
            num_states=ode.num_full_states,
            args=args,
            args_doc=args_doc,
            jac_name=params.functions.jacobian.result_name,
            jacobian_function_name=params.functions.jacobian.function_name,
        )

    if monitored and params.functions.monitored.generate:
        monitor_declaration = monitor_declaration_template.format(
            num_states=ode.num_full_states,
            num_monitored=len(monitored),
            args=args,
            args_doc=args_doc,
            monitored_name=params.functions.monitored.result_name,
            monitored_function_name=params.functions.monitored.function_name,
        )

    pgen = PythonCodeGenerator(python_params)
    cgen = CCodeGenerator(params)

    pcode = "\n\n".join(list(pgen.code_dict(ode, monitored=monitored).values()))

    ccode = "\n\n".join(
        list(
            cgen.code_dict(
                ode, monitored=monitored, include_init=False, include_index_map=False
            ).values()
        )
    )

    # push_log_level(INFO)
    info("Calling GOTRAN just-in-time (JIT) compiler, this may take some " "time...")
    sys.stdout.flush()

    # Configure instant and add additional system headers
    instant_kwargs = configure_instant()

    declaration_form = dict(
        num_states=ode.num_full_states,
        num_parameters=ode.num_parameters,
        num_monitored=len(monitored),
        python_code=pcode,
        args=args,
        args_doc=args_doc,
        jacobian_declaration=jacobian_declaration,
        rhs_name=params.functions.rhs.result_name,
        rhs_function_name=params.functions.rhs.function_name,
        states_name=params.code.states.array_name,
        parameters_name=params.code.parameters.array_name,
        monitor_declaration=monitor_declaration,
    )

    additional_declarations = (
        additional_declarations_
        if additional_declarations is None
        else additional_declarations
    )
    # Compile extension module with instant
    compiled_module = instant.build_module(
        code=ccode,
        additional_declarations=additional_declarations.format(**declaration_form),
        signature=modulename,
        **instant_kwargs
    )

    info(" done")
    # pop_log_level()
    sys.stdout.flush()
    return compiled_module


def configure_instant():
    """
    Check system requirements

    Returns a dict with kwargs that can be passed to instant.build_module.
    """
    instant_kwargs = {}
    swig_include_dirs = []

    instant_kwargs["swig_include_dirs"] = swig_include_dirs
    instant_kwargs["include_dirs"] = [numpy.get_include()]
    instant_kwargs["system_headers"] = [
        "numpy/arrayobject.h",
        "math.h",
    ]  # , "complex.h"]
    instant_kwargs["swigargs"] = ["-O -c++"]
    instant_kwargs["cppargs"] = ["-O2"]

    return instant_kwargs



#!/usr/bin/env python3

import os

from gotran.model.loadmodel import load_ode
from gotran.codegeneration.codegenerators import JuliaCodeGenerator
from gotran.common.options import parameters
from gotran.common import error, info, list_timings, Timer
from modelparameters.parameterdict import *


def gotran2julia(filename, params):
    """
    Create a c header file from a gotran model
    """
    timer = Timer("Generate Julia code from {}".format(filename))

    # Load Gotran model
    ode = load_ode(filename)

    # Collect monitored
    if params.functions.monitored.generate:
        monitored = [expr.name for expr in ode.intermediates + ode.state_expressions]
    else:
        monitored = None

    # Create a C code generator
    gen = JuliaCodeGenerator(params)

    output = params.output

    if output:
        if not output.endswith(".jl"):
            output += ".jl"
    else:
        output = filename.replace(".ode", "") + ".jl"

    info("")
    info("Generating Julia code for the {0} ode...".format(ode.name))
    code = gen.module_code(ode, monitored=monitored)
    info("  done.")
    with open(output, "w") as f:
        f.write(code)

    del timer

    if params.list_timings:
        list_timings()


def main():
    import sys, os

    params = ParameterDict(
        list_timings=Param(
            False,
            description="If true timings for reading "
            "and evaluating the model is listed.",
        ),
        output=Param("", description="Specify output file name"),
        **JuliaCodeGenerator.default_parameters()
    )

    params.parse_args(usage="usage: %prog FILE [options]")

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotran2julia(file_name, params)


if __name__ == "__main__":
    main()



# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

from . import symbolicnewtonsolution

# gotran imports
from gotran.algorithms.symbolicnewtonsolution import *

__all__ = [_name for _name in list(globals().keys()) if _name[0] != "_"]

# del _name



# Copyright (C) 2013 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    "ExplicitEuler",
    "explicit_euler_solver",
    "RushLarsen",
    "rush_larsen_solver",
    "GeneralizedRushLarsen",
    "generalized_rush_larsen_solver",
    "HybridGeneralizedRushLarsen",
    "hybrid_generalized_rush_larsen_solver",
    "SimplifiedImplicitEuler",
    "simplified_implicit_euler_solver",
    "get_solver_fn",
]

# System imports
import sys

# ModelParameters imports
from modelparameters.sympytools import sp, Conditional
from modelparameters.codegeneration import sympycode

# Local imports
from gotran.common import (
    error,
    info,
    debug,
    check_arg,
    check_kwarg,
    scalars,
    Timer,
    warning,
    tuplewrap,
    parameters,
    warning,
)
from gotran.common import listwrap
from gotran.model.utils import ode_primitives
from gotran.model.odeobjects import State, Parameter, IndexedObject, Comment
from gotran.model.expressions import *
from gotran.model.ode import ODE
from gotran.codegeneration.codecomponent import CodeComponent


def explicit_euler_solver(ode, function_name="forward_explicit_euler", params=None):
    """
    Return an ODEComponent holding expressions for the explicit Euler method

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return ExplicitEuler(ode, function_name=function_name, params=params)


def rush_larsen_solver(
    ode, function_name="forward_rush_larsen", delta=1e-8, params=None
):
    """
    Return an ODEComponent holding expressions for the Rush Larsen method

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    delta : float
        Value to safeguard the evaluation of the rush larsen step.
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return RushLarsen(ode, function_name=function_name, delta=delta, params=params)


def generalized_rush_larsen_solver(
    ode, function_name="forward_generalized_rush_larsen", delta=1e-8, params=None
):
    """
    Return an ODEComponent holding expressions for the generalized Rush Larsen method

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    delta : float
        Value to safeguard the evaluation of the rush larsen step.
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return GeneralizedRushLarsen(
        ode, function_name=function_name, delta=delta, params=params
    )


def hybrid_generalized_rush_larsen_solver(
    ode,
    function_name="forward_hybrid_generalized_rush_larsen",
    delta=1e-8,
    params=None,
    stiff_states=None,
):
    """
    Return an ODEComponent holding expressions for the generalized Rush Larsen method

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    delta : float
        Value to safeguard the evaluation of the rush larsen step.
    params : dict
        Parameters determining how the code should be generated
    stiff_state_variables : list of str
        States that are stiff and should be solved with GRL1. The remaining states are solved with explicit euler
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return HybridGeneralizedRushLarsen(
        ode,
        function_name=function_name,
        delta=delta,
        params=params,
        stiff_state_variables=stiff_states,
    )


def simplified_implicit_euler_solver(
    ode,
    function_name="forward_simplified_implicit_euler",
    numeric_jacobian=False,
    params=None,
):
    """
    Return an ODEComponent holding expressions for the simplified
    implicit Euler method

    Arguments
    ---------
    ode : gotran.ODE
        The ODE for which the jacobian expressions should be computed
    function_name : str
        The name of the function which should be generated
    numeric_jacobian : bool
        If True use numeric calculated diagonal jacobian.
    params : dict
        Parameters determining how the code should be generated
    """
    if not ode.is_finalized:
        error("The ODE is not finalized")

    return SimplifiedImplicitEuler(
        ode,
        function_name=function_name,
        numeric_jacobian=numeric_jacobian,
        params=params,
    )


def get_solver_fn(solver_type):
    return {
        "explicit_euler": explicit_euler_solver,
        "rush_larsen": rush_larsen_solver,
        "generalized_rush_larsen": generalized_rush_larsen_solver,
        "hybrid_generalized_rush_larsen": hybrid_generalized_rush_larsen_solver,
        "simplified_implicit_euler": simplified_implicit_euler_solver,
    }[solver_type]


class ExplicitEuler(CodeComponent):
    """
    An ODEComponent which compute one step of the explicit Euler algorithm
    """

    def __init__(self, ode, function_name="forward_explicit_euler", params=None):
        """
        Create an ExplicitEuler CodeComponent

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(ode, ODE)

        if ode.is_dae:
            error("Cannot generate an explicit Euler forward step for a DAE.")

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the explicit Euler algorithm to the "
            "{0} ODE".format(ode)
        )
        super(ExplicitEuler, self).__init__(
            "ExplicitEuler",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt"],
        )

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explicit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states
        result_name = self._params.states.array_name
        state_offset = self._params.states.add_offset
        self.shapes[result_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):
            dependent = expr if recount else None
            self.add_indexed_expression(
                result_name,
                (i,),
                expr.state.sym + dt * expr.sym,
                offset_str,
                dependent=dependent,
                enum=expr.state,
            )

        # Call recreate body with the solver expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class RushLarsen(CodeComponent):
    """
    An ODEComponent which compute one step of the Rush Larsen algorithm
    """

    def __init__(
        self, ode, function_name="forward_rush_larsen", delta=1e-8, params=None
    ):
        """
        Create a RushLarsen Solver component

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        delta : float
            Value to safeguard the evaluation of the rush larsen step.
        params : dict
            Parameters determining how the code should be generated
        """

        timer = Timer("Create RushLarsen expressions")
        check_arg(ode, ODE)

        if ode.is_dae:
            error("Cannot generate a Rush Larsen forward step for a DAE.")

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the rush larsen algorithm to the "
            "{0} ODE".format(ode)
        )
        super(RushLarsen, self).__init__(
            "RushLarsen",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt"],
        )

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explcit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states

        result_name = self._params.states.array_name
        state_offset = self._params.states.add_offset
        self.shapes[result_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        might_take_time = len(states) >= 10

        if might_take_time:
            info(
                "Calculating derivatives of {0}. Might take some time...".format(
                    ode.name
                )
            )
            sys.stdout.flush()

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):

            dependent = expr if recount else None

            # Diagonal jacobian value
            time_diff = Timer("Differentiate state_expressions for RushLarsen")
            expr_diff = expr.expr.diff(expr.state.sym)
            del time_diff

            # print expr.state.sym, expr_diff, expr_diff.args
            if expr_diff and expr.state.sym not in expr_diff.args:

                linearized = self.add_intermediate(
                    expr.name + "_linearized", expr_diff, dependent=dependent
                )

                # Solve "exact" using exp
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    expr.state.sym
                    + Conditional(
                        abs(linearized) > delta,
                        expr.sym / linearized * (sp.exp(linearized * dt) - 1.0),
                        dt * expr.sym,
                    ),
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )

            else:

                # Explicit Euler step
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    expr.state.sym + dt * expr.sym,
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )

        if might_take_time:
            info(" done")

        # Call recreate body with the solver expressions as the result
        # expressions
        del timer
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class RushLarsenOneStep(CodeComponent):
    """
    An ODEComponent which compute one step of the Rush Larsen algorithm
    """

    def __init__(self, ode, function_name="forward_rush_larsen", params=None):
        """
        Create a RushLarsen Solver component

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(ode, ODE)

        if ode.is_dae:
            error("Cannot generate a Rush Larsen forward step for a DAE.")

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the rush larsen algorithm to the "
            "{0} ODE".format(ode)
        )
        state_name = params.states.array_name
        super(RushLarsen, self).__init__(
            "RushLarsen",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt", state_name + "_0"],
        )

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explcit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states

        result_name = self._params.states.array_name + "_1"
        previous_name = self._params.states.array_name + "_0"
        state_offset = self._params.states.add_offset
        self.shapes[result_name] = (len(states),)
        self.shapes[previous_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):

            prev = self.add_indexed_object(previous_name, (i,), offset_str)

            dependent = expr if recount else None

            # Diagonal jacobian value
            expr_diff = expr.expr.diff(expr.state.sym)

            if expr_diff and expr.state.sym not in expr_diff:

                linearized = self.add_intermediate(
                    expr.name + "_linearized", expr_diff, dependent=dependent
                )

                # Solve "exact" using exp
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    prev + expr.sym / linearized * (sp.exp(linearized * dt) - 1.0),
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )

            else:

                # Explicit Euler step
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    prev + dt * expr.sym,
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )

        # Call recreate body with the solver expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class GeneralizedRushLarsen(CodeComponent):
    """
    An ODEComponent which compute one step of the Generalized Rush Larsen (GRL1) algorithm
    """

    def __init__(
        self,
        ode,
        function_name="forward_generalized_rush_larsen",
        delta=1e-8,
        params=None,
    ):
        """
        Create a GeneralizedRushLarsen Solver component

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        delta : float
            Value to safeguard the evaluation of the rush larsen step.
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(ode, ODE)

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the rush larsen algorithm to the "
            "{0} ODE".format(ode)
        )
        super(GeneralizedRushLarsen, self).__init__(
            "GeneralizedRushLarsen",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt"],
        )

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explcit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states
        result_name = self._params.states.array_name
        state_offset = self._params.states.add_offset

        self.shapes[result_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):

            expr_diff = expr.expr.diff(expr.state.sym)
            dependent = expr if recount else None
            if expr_diff.is_zero:
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    expr.state.sym + dt * expr.sym,
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )
                continue

            linearized = self.add_intermediate(
                expr.name + "_linearized", expr_diff, dependent=dependent
            )

            # Solve "exact" using exp
            self.add_indexed_expression(
                result_name,
                (i,),
                expr.state.sym
                + Conditional(
                    abs(linearized) > delta,
                    expr.sym / linearized * (sp.exp(linearized * dt) - 1.0),
                    dt * expr.sym,
                ),
                offset_str,
                dependent=dependent,
                enum=expr.state,
            )

        # Call recreate body with the solver expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class HybridGeneralizedRushLarsen(CodeComponent):
    """
    An ODEComponent which compute one step of the hybrid explicit Euler / Generalized Rush Larsen (GRL1) scheme
    """

    def __init__(
        self,
        ode,
        function_name="forward_hybrid_generalized_rush_larsen",
        delta=1e-8,
        params=None,
        stiff_state_variables=None,
    ):
        """
        Create a HybridGeneralizedRushLarsen Solver component

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        delta : float
            Value to safeguard the evaluation of the rush larsen step.
        params : dict
            Parameters determining how the code should be generated

        """
        check_arg(ode, ODE)

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the rush larsen algorithm to the "
            "{0} ODE".format(ode)
        )
        super(HybridGeneralizedRushLarsen, self).__init__(
            "HybridGeneralizedRushLarsen",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt"],
        )

        state_names = [s.name for s in self.root.full_states]
        if stiff_state_variables is None:
            stiff_state_variables = []
        elif type(stiff_state_variables) is str:
            stiff_state_variables = stiff_state_variables.split(",")

        for s in stiff_state_variables:
            if s == "":
                continue
            assert s in state_names, "Unknown state '{}'".format(s)

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explcit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states
        result_name = self._params.states.array_name
        state_offset = self._params.states.add_offset

        self.shapes[result_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):
            state_is_stiff = state_names[i] in stiff_state_variables

            expr_diff = expr.expr.diff(expr.state.sym)
            dependent = expr if recount else None
            if not state_is_stiff or expr_diff.is_zero:
                self.add_indexed_expression(
                    result_name,
                    (i,),
                    expr.state.sym + dt * expr.sym,
                    offset_str,
                    dependent=dependent,
                    enum=expr.state,
                )
                continue

            linearized = self.add_intermediate(
                expr.name + "_linearized", expr_diff, dependent=dependent
            )

            # Solve "exact" using exp
            self.add_indexed_expression(
                result_name,
                (i,),
                expr.state.sym
                + Conditional(
                    abs(linearized) > delta,
                    expr.sym / linearized * (sp.exp(linearized * dt) - 1.0),
                    dt * expr.sym,
                ),
                offset_str,
                dependent=dependent,
                enum=expr.state,
            )

        # Call recreate body with the solver expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)


class SimplifiedImplicitEuler(CodeComponent):
    """
    An ODEComponent which compute one step of a simplified Implicit Euler
    algorithm
    """

    def __init__(
        self,
        ode,
        function_name="forward_simplified_implicit_euler",
        numeric_jacobian=False,
        params=None,
    ):
        """
        Create a SimplifiedImplicitEuler CodeComponent

        Arguments
        ---------
        ode : gotran.ODE
            The parent component of this ODEComponent
        function_name : str
            The name of the function which should be generated
        numeric_jacobian : bool
            If True use numeric calculated diagonal jacobian.
        params : dict
            Parameters determining how the code should be generated
        """
        check_arg(ode, ODE)

        if ode.is_dae:
            error("Cannot generate an explicit Euler forward step for a DAE.")

        # Call base class using empty result_expressions
        descr = (
            "Compute a forward step using the simplified implicit Euler"
            "algorithm to the {0} ODE".format(ode)
        )
        super(SimplifiedImplicitEuler, self).__init__(
            "SimplifiedImplicitEuler",
            ode,
            function_name,
            descr,
            params=params,
            additional_arguments=["dt"],
        )

        # Recount the expressions if representation of states are "array" as
        # then the method is not full explcit
        recount = self._params.states.representation != "array"

        # Gather state expressions and states
        state_exprs = self.root.state_expressions
        states = self.root.full_states

        result_name = self._params.states.array_name
        state_offset = self._params.states.add_offset
        self.shapes[result_name] = (len(states),)

        # Get time step and start creating the update algorithm
        if self._params.states.add_offset:
            offset_str = "{0}_offset".format(result_name)
        else:
            offset_str = ""

        dt = self.root._dt.sym
        for i, expr in enumerate(state_exprs):

            dependent = expr if recount else None

            # Diagonal jacobian value
            diag_jac_expr = expr.expr.diff(expr.state.sym)

            if not diag_jac_expr.is_zero:
                diag_jac = self.add_intermediate(
                    expr.name + "_diag_jac", diag_jac_expr, dependent=dependent
                )
            else:
                diag_jac = 0.0

            # Add simplified single Implicit Euler step
            self.add_indexed_expression(
                result_name,
                (i,),
                expr.state.sym + dt * expr.sym / (1 - dt * diag_jac),
                offset_str,
                enum=expr.state,
            )

        # Call recreate body with the solver expressions as the result
        # expressions
        results = {result_name: self.indexed_objects(result_name)}
        results, body_expressions = self._body_from_results(**results)
        self.body_expressions = self._recreate_body(body_expressions, **results)



"""
To use install assimulo which is a python wrapper of
the sundials solvers

conda install assimulo

"""
from .utils import suppress_stdout_stderr

# Assimulo imports
try:
    with suppress_stdout_stderr():
        from assimulo.solvers import CVode
        from assimulo.problem import Explicit_Problem
    has_sundials = True
except ImportError as ex:
    has_sundials = False
from .odesolver import Solver, ODESolverError

# Local imports
__all__ = ["SundialsSolver", "has_sundials", "SundialsNotInstalled"]


class SundialsNotInstalled(Exception):
    pass


class SundialsSolver(Solver):
    def __init__(self, ode, method="cvode", **options):

        # Check imports
        if not has_sundials:
            msg = "Chosen backend is sundials, but sundials is " + "not installed"
            raise SundialsNotInstalled(msg)

        # # Check method
        # msg = ("Method {} is not a sundials method. ".format(method)+\
        #        "Possible methods are {}".format(sundials_methods))
        # assert method in sundials_methods, msg
        # self._method = method

        self._options = options

        Solver.__init__(
            self,
            ode,
            arguments="tsp",
            additional_declarations=additional_declarations,
            jacobian_declaration_template=jacobian_declaration_template,
            **options
        )
        self._create_solver()

    def _create_solver(self):
        # Create problem
        self._problem = Explicit_Problem(self._rhs, self._y0)

        # Set Jacobian if used
        if self._jac is not None:
            self._problem.jac = self._jac
            self._options["usejac"] = True

        # Create the solver
        # if self._method == "cvode":
        self._solver = CVode(self._problem)

        # Parse parameters to rhs
        self._solver.sw = self._model_params.tolist()
        self._solver.problem_info["switches"] = True

        # Set verbosity to 100 (i.e turn of printing) if not specified
        verbosity = self._options.pop("verbosity", 100)
        self._options["verbosity"] = verbosity

        # Set verbosity to 100 (i.e turn of printing) if not specified
        maxh = self._options.pop("maxh", 5e-4)
        self._options["maxh"] = maxh

        self._solver.options.update(
            (k, v)
            for k, v in self._options.items()
            if k in list(self._solver.options.keys())
        )

    def _eval_monitored(self, time, res, params, values):
        self.module.monitor(time, res, params, values)

    @property
    def problem(self):
        return self._problem

    @property
    def solver(self):
        return self._solver

    def get_options(self):
        """
        Get solver options for the current solver
        """
        return self.solver.options

    @staticmethod
    def list_solver_options():

        # if method == "cvode":
        return _CVode.default_options()

    def _solve(self, tsteps, *args, **kwargs):
        """
        Solve the problem
        """

        self._create_solver()

        t_end = tsteps[-1]
        ncp = len(tsteps)
        ncp_list = tsteps
        # Construct problem
        t, y = self._solver.simulate(t_end, ncp, ncp_list)
        # t,y = self._solver.simulate(t_end)

        return t, y


if has_sundials:

    class _CVode:
        @staticmethod
        def default_options():
            d = {
                "atol": np.array([]),
                "backward": False,
                "clock_step": False,
                "discr": "BDF",
                "display_progress": True,
                "dqrhomax": 0.0,
                "dqtype": "CENTERED",
                "external_event_detection": False,
                "inith": 0.0,
                "iter": "Newton",
                "linear_solver": "DENSE",
                "maxcor": 3,
                "maxcorS": 3,
                "maxh": 0.0,
                "maxkrylov": 5,
                "maxncf": 10,
                "maxnef": 20,
                "maxord": 5,
                "maxsteps": 500,
                "minh": 0.0,
                "nnz": -1,
                "norm": "WRMS",
                "num_threads": 1,
                "pbar": [],
                "precond": "Banded",
                "report_continuously": False,
                "rtol": 1e-06,
                "sensmethod": "STAGGERED",
                "stablimit": False,
                "store_event_points": True,
                "suppress_sens": False,
                "time_limit": 0,
                "usejac": False,
                "usesens": False,
                "verbosity": 30,
            }
            d.pop("atol")
            return d


additional_declarations = r"""
%init%{{
import_array();
%}}

%include <exception.i>
%feature("autodoc", "1");

// Typemaps
%typemap(in) (double* {rhs_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles "
                  "expected. Make sure the numpy array is contiguous, "
                  "and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                                  "{num_states}, for the {rhs_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (const double* {states_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                              "{num_states}, for the {states_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (double* {states_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                              "{num_states}, for the {states_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%typemap(in) (double* {parameters_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles expected."
           " Make sure the numpy array is contiguous, and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_parameters} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
             "{num_parameters}, for the {parameters_name} argument.");

  $1 = (double *)PyArray_DATA(xa);

}}

// The typecheck
%typecheck(SWIG_TYPECHECK_DOUBLE_ARRAY) double *
{{
    $1 = PyArray_Check($input) ? 1 : 0;
}}

%typecheck(SWIG_TYPECHECK_DOUBLE_ARRAY) const double *
{{
    $1 = PyArray_Check($input) ? 1 : 0;
}}

%pythoncode%{{
def {rhs_function_name}({args}):
    '''
    Evaluates the right hand side of the model

    Arguments:
    ----------
{args_doc}
    '''
    import numpy as np

    {rhs_name} = np.zeros_like({states_name})

    if not isinstance(parameters, np.ndarray):
        parameters = np.array(parameters, dtype=np.float_)
    _{rhs_function_name}({args}, {rhs_name})
    return {rhs_name}

{python_code}
%}}

%rename (_{rhs_function_name}) {rhs_function_name};

{jacobian_declaration}

{monitor_declaration}
"""


jacobian_declaration_template = """
// Typemaps
%typemap(in) (double* {jac_name})
{{
  // Check type
  if (!PyArray_Check($input))
    SWIG_exception(SWIG_TypeError, "Numpy array expected");

  // Get PyArrayObject
  PyArrayObject *xa = (PyArrayObject *)$input;

  // Check data type
  if (!(PyArray_ISCONTIGUOUS(xa) && PyArray_TYPE(xa) == NPY_DOUBLE))
    SWIG_exception(SWIG_TypeError, "Contigous numpy array of doubles "
                  "expected. Make sure the numpy array is contiguous, "
                  "and uses dtype=float_.");

  // Check size of passed array
  if ( PyArray_SIZE(xa) != {num_states}*{num_states} )
    SWIG_exception(SWIG_ValueError, "Expected a numpy array of size: "
                                  "{num_states}*{num_states}, for the {jac_name} argument.");

  $1 = (double *)PyArray_DATA(xa);
}}

%pythoncode%{{
def {jacobian_function_name}({args}, {jac_name}=None):
    '''
    Evaluates the jacobian of the model

    Arguments:
    ----------
{args_doc}
    {jac_name} : np.ndarray (optional)
        The computed result
    '''
    import numpy as np

    {jac_name} = np.zeros({num_states}*{num_states}, dtype=np.float_)

    _{jacobian_function_name}({args}, {jac_name})
    {jac_name}.shape = ({num_states},{num_states})
    return {jac_name}
%}}

%rename (_{jacobian_function_name}) {jacobian_function_name};
"""



#!/usr/bin/env python3

import numpy as np
import os

from gotran.codegeneration.codegenerators import DOLFINCodeGenerator
from gotran.model.loadmodel import load_ode
from gotran.codegeneration.algorithmcomponents import rhs_expressions


def gotran2dolfin(filename, params):
    """
    Create ufl code from a gotran model
    """

    # Load Gotran model
    ode = load_ode(filename)

    # Create a ufl based model code generator
    code_gen = DOLFINCodeGenerator(params)

    output = params.output

    if output:
        if not output.endswith(".py"):
            output += ".py"
    else:
        output = filename.replace(".ode", "") + ".py"

    f = open(output, "w")
    f.write("from __future__ import division\n\n")
    f.write(code_gen.init_states_code(ode) + "\n\n")
    f.write(code_gen.init_parameters_code(ode) + "\n\n")
    f.write(
        code_gen.function_code(rhs_expressions(ode, params=code_gen.params.code)) + "\n"
    )


def main():
    import sys
    from modelparameters.parameterdict import ParameterDict, OptionParam, Param

    params = ParameterDict(
        output=Param("", description="Specify output file name"),
        **DOLFINCodeGenerator.default_parameters()
    )
    params.parse_args(usage="usage: %prog FILE [options]")  # sys.argv[2:])

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotran2dolfin(file_name, params)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2013-05-07 -- 2015-06-04"
__copyright__ = "Copyright (C) 2013 " + __author__
__license__ = "GNU LGPL Version 3.0 or later"

from gotran import *
from gotran.model.loadmodel import load_ode
from gotran.model.odeobjects import Comment
from gotran.common import error
from modelparameters.parameterdict import *


def gotranexport(filename, params):

    if len(params.components) == 1 and params.components[0] == "":
        error("Expected at least 1 component.")

    if params.name == "":
        error("Expected a name of the exported ODE.")

    if params.name in params.components:
        error("Expected the name of the ODE to be different from the components.")

    sub_ode = ODE(params.name)
    sub_ode.import_ode(load_ode(filename), components=params.components)
    sub_ode.finalize()
    sub_ode.save()


def main():
    import sys, os

    params = ParameterDict(
        components=Param(
            [""], description="List all components that will be " "exported."
        ),
        name=Param("", description="Specify name of exported ODE."),
    )
    params.parse_args(usage="usage: %prog FILE [options]")

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotranexport(file_name, params)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# System imports
from setuptools import setup, Command
from shutil import rmtree
from pathlib import Path
import os
import io
from os.path import join as pjoin
import glob
import platform
import sys

# Version number
major = 2021
minor = 0.1
VERSION = "{0}.{1}".format(major, minor)

DESCRIPTION = "A declarative language describing ordinary differential equations."

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


scripts = [
    os.path.splitext(script)[0]
    for script in glob.glob("gotran/scripts/*")
    if Path(script).suffix == ".py" and Path(script).stem != "__init__"
]

requirements = [
    "sympy<=1.1.1",
    "numpy",
    "scipy",
    "matplotlib",
    "networkx",
    "six",
    "future",
    "modelparameters",
    "instant2",
]

if platform.system() == "Windows" or "bdist_wininst" in sys.argv:
    # In the Windows command prompt we can't execute Python scripts
    # without a .py extension. A solution is to create batch files
    # that runs the different scripts.
    batch_files = []
    for script in scripts:
        batch_file = script + ".bat"
        with open(batch_file, "w") as f:
            f.write('python "%%~dp0\%s" %%*\n' % os.path.split(script)[1])
        batch_files.append(batch_file)
    scripts.extend(batch_files)


class clean(Command):
    """
    Cleans *.pyc so you should get the same copy as is in the VCS.
    """

    description = "remove build files"
    user_options = [("all", "a", "the same")]

    def initialize_options(self):
        self.all = None

    def finalize_options(self):
        pass

    def run(self):
        import os

        os.system("utils/clean-files")


class run_tests(Command):
    """
    Runs all tests under the modelparameters/ folder
    """

    description = "run all tests"
    user_options = []  # distutils complains if this is not here.

    def __init__(self, *args):
        self.args = args[0]  # so we can pass it to other classes
        Command.__init__(self, *args)

    def initialize_options(self):  # distutils wants this
        pass

    def finalize_options(self):  # this too
        pass

    def run(self):
        import os

        os.system("python utils/run_tests.py")


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("python -m twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


setup(
    name="gotran",
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ComputationalPhysiology/gotran",
    author="Johan Hake and Henrik Finsberg",
    author_email="henriknf@simula.no",
    packages=[
        "gotran",
        "gotran.common",
        "gotran.model",
        "gotran.algorithms",
        "gotran.codegeneration",
        "gotran.input",
        "gotran.solver",
        "gotran.scripts",
    ],
    install_requires=requirements,
    entry_points={
        "console_scripts": ["gotran=gotran.__main__:main"]
        + [
            f"{Path(script).stem}={'.'.join(script.split(os.path.sep))}:main"
            for script in scripts
        ]
    },
    cmdclass={"test": run_tests, "clean": clean, "upload": UploadCommand},
)



# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["SymbolicNewtonSolution"]

# System imports
from modelparameters.sympytools import sp
from modelparameters.utils import check_arg, scalars

# Local imports
from gotran.model.ode import ODE
from gotran.common.dicts import odict
from collections import OrderedDict
from functools import reduce


def _iszero(x):
    """Returns True if x is zero."""
    x = sp.sympify(x)
    return x.is_zero


class SymbolicNewtonSolution(object):
    """
    Class for storing information about a symbolic solution of newton
    iteration of an ODE
    """

    def __init__(self, ode, theta=1):
        """
        Create a SymbolicNewtonSolution

        Arguments
        ---------
        ode : ODE
            The ODE which the symbolc solution is created for
        theta : scalar \xe2\x88\x88 [0,1]
            Theta for creating the numerical integration rule of the ODE
        """

        check_arg(ode, ODE, 0, SymbolicNewtonSolution)
        check_arg(theta, scalars, 1, SymbolicNewtonSolution, ge=0, le=1)

        # Store attributes
        self.ode = ode
        self.theta = theta

        # Create symbolic linear system
        self.F, self.F_expr, self.jacobi, self.states, self.jac_subs = _create_newton_system(
            ode, theta
        )

        # Create a simplified LU decomposition of the jacobi matrix
        # FIXME: Better names!
        self.x, self.new_old, self.old_new = _LU_solve(self.jacobi, self.F)
        # self.LU_decomp, self.perm, self.LU_decomp_subs_0,\
        #    self.LU_decomp_subs_1 = _LU_solve(self.jacobi, self.F)


def _LU_solve(AA, rhs):
    """
    Returns the symbolic solve of AA*x=rhs
    """
    if not AA.is_square:
        raise NonSquareMatrixError()
    n = AA.rows
    A = AA[:, :]
    p = []

    nnz = 0
    for i in range(n):
        for j in range(n):
            nnz += not _iszero(A[i, j])

    print("Num non zeros in jacobian:", nnz)

    # A map between old symbols and new. The values in this dict corresponds to
    # where an old symbol is used. If the length of the value is 1 it is only
    # used once and once the value is used it can be freed.
    old_new = OrderedDict()
    new_old = OrderedDict()
    new_count = 0
    global zero_operations
    zero_operations = 0

    def update_entry(new_count, i, j, k):
        global zero_operations
        if _iszero(A[i, k] * A[k, j]):
            zero_operations += 1
            return new_count

        # Create new symbol and store the representation
        new_sym = sp.Symbol("j_{0}_{1}:{2}".format(i, j, new_count))
        new_old[new_sym] = A[i, j] - A[i, k] * A[k, j]
        for old_sym in [A[i, j], A[i, k], A[k, j]]:
            storage = old_new.get(old_sym)
            if storage is None:
                storage = set()
                old_new[old_sym] = storage
            storage.add(new_sym)

        # Change entry to the new symbol
        A[i, j] = new_sym
        new_count += 1

        return new_count

    # factorization
    for j in range(n):
        for i in range(j):
            for k in range(i):
                new_count = update_entry(new_count, i, j, k)
        pivot = -1
        for i in range(j, n):
            for k in range(j):
                new_count = update_entry(new_count, i, j, k)

            # find the first non-zero pivot, includes any expression
            if pivot == -1 and not _iszero(A[i, j]):
                pivot = i
        if pivot < 0:
            # this result is based on iszerofunc's analysis of the
            # possible pivots, so even though the element may not be
            # strictly zero, the supplied iszerofunc's evaluation gave
            # True
            raise ValueError("No nonzero pivot found; inversion failed.")

        if pivot != j:  # row must be swapped
            A.row_swap(pivot, j)
            p.append([pivot, j])
        scale = 1 / A[j, j]
        for i in range(j + 1, n):
            if _iszero(A[i, j]):
                zero_operations += 1
                continue

            # Create new symbol and store the representation
            new_sym = sp.Symbol("j_{0}_{1}:{2}".format(i, j, new_count))
            new_old[new_sym] = A[i, j] * scale
            for old_sym in [A[i, j], A[j, j]]:
                storage = old_new.get(old_sym)
                if storage is None:
                    storage = set()
                    old_new[old_sym] = storage
                storage.add(new_sym)

            # Change entry to the new symbol
            A[i, j] = new_sym
            new_count += 1

    nnz = 0
    for i in range(n):
        for j in range(n):
            nnz += not _iszero(A[i, j])

    print("Num non zeros in factorized jacobian:", nnz)
    print("Num non-zero operations while factorizing matrix:", new_count)
    print("Num zero operations while factorizing matrix:", zero_operations)
    factorizing_nnz_operations = new_count
    zero_operations = 0

    n = AA.rows
    b = rhs.permuteFwd(p)

    def update_entry(new_count, i, j):
        global zero_operations
        if _iszero(b[j, 0] * A[i, j]):
            zero_operations += 1
            return new_count

        # Create new symbol and store the representation
        new_sym = sp.Symbol("F_{0}:{1}".format(i, new_count))
        new_old[new_sym] = b[i, 0] - b[j, 0] * A[i, j]
        for old_sym in [b[i, 0], b[j, 0], A[i, j]]:
            storage = old_new.get(old_sym)
            if storage is None:
                storage = set()
                old_new[old_sym] = storage
            storage.add(new_sym)

        # Change entry to the new symbol
        b[i, 0] = new_sym
        new_count += 1

        return new_count

    # forward substitution, all diag entries are scaled to 1
    for i in range(n):
        for j in range(i):
            new_count = update_entry(new_count, i, j)
            # b.row(i, lambda x,k: x - b[j,k]*A[i,j])

    # backward substitution
    for i in range(n - 1, -1, -1):
        for j in range(i + 1, n):
            new_count = update_entry(new_count, i, j)
            # b.row(i, lambda x,k: x - b[j,k]*A[i,j])
        b.row(i, lambda x, k: x / A[i, i])

    print(
        "Num zero operations while forward/backward substituting matrix:",
        zero_operations,
    )
    print(
        "Num non-zero operations while factorizing matrix:",
        new_count - factorizing_nnz_operations,
    )

    return b, new_old, old_new

    return A, p, old_new, new_old


def symbolic_solve(ode, theta=1):
    """
    Symbolically solve an ode system
    """


def _create_newton_system(ode, theta=1):
    """
    Return the nonlinear F, the Jacobian matrix and a mapping of states
    to nonero fields for the ODE
    """

    # Lists of symbols
    states = [state.sym for state in ode.states]
    states_0 = [state.sym_0 for state in ode.states]
    vars_ = [var.sym for var in ode.variables]
    vars_0 = [var.sym_0 for var in ode.variables]

    def sum_ders(ders):
        return reduce(lambda x, y: x + y, ders, 0)

    # Substitution dict between sym and previous sym value
    subs = [syms for syms in zip(states + vars_, states_0 + vars_0)]

    # Generate F using theta rule
    F_expr = [
        theta * expr
        + (1 - theta) * expr.subs(subs)
        - (sum_ders(ders) - sum_ders(ders).subs(subs)) / ode.dt
        for ders, expr in ode.get_derivative_expr(True)
    ]

    # Create mapping of orginal expression
    sym_map = OrderedDict()
    jac_subs = OrderedDict()
    F = []
    for i, expr in enumerate(F_expr):
        for j, state in enumerate(states):
            F_ij = expr.diff(state)
            if F_ij:
                jac_sym = sp.Symbol("j_{0}_{1}".format(i, j))
                sym_map[i, j] = jac_sym
                jac_subs[jac_sym] = F_ij
                # print "[%d,%d] (%d) # [%s, %s]  \n%s" \
                # % (i,j, len(F_ij.args), states[i], states[j], F_ij)

        F_i = sp.Symbol("F_{0}".format(i))
        sym_map[i, j + 1] = F_i
        jac_subs[F_i] = expr
        F.append(F_i)

    # Create the Jacobian
    jacobi = sp.SparseMatrix(
        len(states), len(states), lambda i, j: sym_map.get((i, j), 0)
    )

    # return Symbolic representation of the linear system
    return sp.Matrix(F), F_expr, jacobi, states, jac_subs



#!/usr/bin/env python3

from gotran.model.loadmodel import load_ode
from gotran.codegeneration.codegenerators import MatlabCodeGenerator
from gotran.common.options import parameters
from gotran.common import error, info, list_timings, Timer


def gotran2matlab(filename, params):
    """
    Create a matlab code from a gotran model
    """

    timer = Timer("Generate Matlab code from {}".format(filename))

    # Load Gotran model
    ode = load_ode(filename)

    gen = MatlabCodeGenerator(params)

    info("")
    info("Generating Matlab files for the {0} ode...".format(ode.name))

    # Collect monitored
    if params.functions.monitored.generate:
        monitored = [expr.name for expr in ode.intermediates + ode.state_expressions]
    else:
        monitored = None

    for function_name, code in list(gen.code_dict(ode, monitored).items()):
        open("{0}_{1}.m".format(ode.name, function_name), "w").write(code)

    info("  done.")


def main():
    import os
    import sys

    from modelparameters.parameterdict import ParameterDict, Param

    generation_params = MatlabCodeGenerator.default_parameters()

    params = ParameterDict(
        list_timings=Param(
            False,
            description="If true timings for reading "
            "and evaluating the model is listed.",
        ),
        output=Param("", description="Specify output file name"),
        **generation_params
    )
    params.parse_args(usage="usage: %prog FILE [options]")  # sys.argv[2:])

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotran2matlab(file_name, params)


if __name__ == "__main__":
    main()



# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["load", "save", "present_time_str"]

# System imports
import time
import os
import numpy

from pickle import Unpickler, Pickler
from os.path import isdir, join

# gotran imports
from gotran.common import *

TIME_FORMAT = "%Y.%m.%d-%H.%M.%S"


def present_time_str():
    "Returns the present time nicely formated"
    return time.strftime(TIME_FORMAT)


def save(basename, **data):
    """
    Save data using cPickle

    @type basename : str
    @param basename : The name of the file to save the data in, .cpickle
    will be appended to the file name if not provided
    @param data : The actuall data to be saved.

    """

    check_arg(basename, str, 0)

    # If zero data size just return
    if len(data) == 0:
        return

    filename = basename if ".cpickle" in basename else basename + ".cpickle"

    f = open(filename, "w")

    p = Pickler(f)

    # Dump the dictionary kwarg
    p.dump(data)
    f.close()


def compare_dicts(p1, p2):
    "Recursively compares a dict of values"
    assert isinstance(p1, dict)
    assert isinstance(p2, dict)
    try:
        ret = p1 == p2
    except ValueError as e:
        ret = True
        for key, value in p1.items():
            if isinstance(value, numpy.ndarray):
                ret = ret and (value == p2[key]).all()
            elif isinstance(value, dict):
                ret = ret and compare_dicts(value, p2[key])
            else:
                ret = ret and value == p2[key]
    return ret


def load(basename, latest_timestamp=False, collect=False):
    """
    Load data using cPickle

    @type basename : str
    @param basename : The name of the file where the data will be loaded from,
    '.cpickle' will be appended to the file name if not provided
    @type latest_timestamp : bool
    @param latest_timestamp : If true return the data from latest version of
    saved data with the same basename
    @type collect : bool
    @param collect : If True collect all data with the same basename and
    the same parameters
    """

    check_arg(basename, str, 0)

    if latest_timestamp and collect:
        raise TypeError("'collect' and 'latest_timestamp' cannot both be True")

    # If not collect just return the data froma single data file
    if not collect:
        return load_single_data(basename, latest_timestamp)

    filenames = get_data_filenames(basename)

    # No filenames with timestamp. Try to return data file without timestamp
    if not filenames:
        return load_single_data(basename, False)

    # Start with the latest filename and load the data and collect them if
    # the data have the same parameter
    data = load_single_data(filenames.pop(-1), False)
    params = data["params"]
    for filename in reversed(filenames):
        local_data = load_single_data(filename, False)
        if not compare_dicts(params, local_data["params"]):
            info("Not the same parameters, skipping data from '%s'", filename)
            continue
        merge_data_dicts(data, local_data)

    return data


def merge_data_dicts(data0, data1):
    "Merge data from data1 into data0"

    def recursively_merge_data(data0, data1):
        for (key, values), org_values in zip(iter(data1.items()), list(data0.values())):
            if isinstance(values, dict):
                data0[key] = recursively_merge_data(org_values, values)
            elif isinstance(values, list):
                if isinstance(values[0], list):
                    for i in range(len(values)):
                        org_values[i].extend(values[i])
                else:
                    org_values.extend(values)
                data0[key] = org_values
        return data0

    for (key, values), org_values in zip(iter(data1.items()), list(data0.values())):
        if key == "params":
            continue

        data0[key] = recursively_merge_data(values, org_values)


def load_single_data(basename, latest_timestamp):
    "Helper function for load"

    if latest_timestamp:
        filenames = get_data_filenames(basename)
        if not filenames:
            raise IOError(
                "No files with timestamp for basename: " "'%s' excist" % basename
            )
        basename = filenames[-1]

    filename = basename if ".cpickle" in basename else basename + ".cpickle"

    if not os.path.isfile(filename):
        raise IOError("No file with basename: '%s' excists" % basename)

    info("Loading data from: %s", filename)
    f = open(filename, "r")

    return Unpickler(f).load()


def get_data_filenames(basename):
    "Helper functions for getting data filenames"
    import glob, re

    basename = basename if ".cpickle" in basename else basename.replace(".cpickle", "")
    pattern = re.compile(
        "%s-[0-9]+\.[0-9]+\.[0-9]+-[0-9]+" "\.[0-9]+\.[0-9]+.cpickle" % basename
    )

    filenames = [
        filename
        for filename in glob.glob("%s*.cpickle" % basename)
        if re.search(pattern, filename)
    ]

    if not filenames:
        return []

    filenames.sort()
    return filenames
